#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase

import mock

from ycyc.tools import rescuekit


class TestRescue(TestCase):
    def test_usage(self):
        self.assertIsInstance(
            rescuekit.Rescue.catch(""),  # pylint: disable=E1120
            rescuekit.Rescue
        )
        self.assertIsInstance(
            rescuekit.Rescue.meet(""),  # pylint: disable=E1120
            rescuekit.Rescue
        )

        func_mock = mock.MagicMock()
        rescue = rescuekit.Rescue.catch("")  # pylint: disable=E1120
        func_mock.return_value = 0
        self.assertEqual(rescue.call(func_mock, 1, val=2), 0)
        func_mock.assert_called_once_with(1, val=2)

        func_mock = mock.MagicMock()
        func_mock.side_effect = ValueError
        self.assertEqual(rescue.call(func_mock, val=3), "")
        func_mock.assert_called_once_with(val=3)

        rescue = rescuekit.Rescue.catch(
            "ValueError", ValueError
        ).catch(
            "KeyError", KeyError
        ).meet(
            "", None
        )
        func_mock = mock.MagicMock()
        func_mock.side_effect = ValueError
        self.assertEqual(rescue.call(func_mock), "ValueError")

        func_mock = mock.MagicMock()
        func_mock.side_effect = KeyError
        self.assertEqual(rescue.call(func_mock), "KeyError")

        func_mock = mock.MagicMock()
        func_mock.return_value = None
        self.assertEqual(rescue.call(func_mock), "")

        self.assertEqual(
            rescuekit.Rescue.catch("nan").call(lambda: 1 / 0),  # pylint: disable=E1120
            "nan"
        )



#!/usr/bin/env python
# encoding: utf-8

import time

from unittest import TestCase

from ycyc.tools import stopwatch


class TestStopwatch(TestCase):
    def test_usage(self):
        s_watch = stopwatch.Stopwatch()

        self.assertAlmostEqual(s_watch.duration, 0)
        self.assertIsNone(s_watch.start_on)

        s_watch.enable = True
        start_on1 = time.time()
        self.assertIsNotNone(s_watch.start_on)
        self.assertTrue(s_watch.enable)
        self.assertGreater(s_watch.duration, 0)

        start_on2 = s_watch.start_on
        duration1 = s_watch.duration
        s_watch.enable = False
        duration2 = s_watch.duration
        self.assertEqual(s_watch.start_on, start_on2)
        self.assertFalse(s_watch.enable)
        self.assertEqual(s_watch.duration, duration2)

        duration3 = time.time() - start_on1
        s_watch.enable = True
        self.assertLess(s_watch.duration, duration3)
        self.assertGreater(s_watch.duration, duration1)
        self.assertGreater(s_watch.start_on, start_on2)



import heapq


class SeqHeapItem(object):

    def __init__(self, sequence, cmp_func):
        self.sequence = iter(sequence)
        self.cmp_func = cmp_func
        self.value = None
        self.backward()

    def backward(self):
        self.value = next(self.sequence)

    def __cmp__(self, other):
        return self.cmp_func(self.value, other.value)


def alignment_seq(sequences, key_func=None, cmp_func=None):
    key_func = key_func or (lambda x: x)
    cmp_func = cmp_func or (lambda x, y: cmp(key_func(x), key_func(y)))

    seq_heap = [SeqHeapItem(i, cmp_func) for i in sequences]

    while seq_heap:
        try:
            heapq.heapify(seq_heap)
            min_item = heapq.nsmallest(1, seq_heap)[0]
            max_item = heapq.nlargest(1, seq_heap)[0]

            if min_item == max_item:
                yield [i.value for i in seq_heap]
                backward_seq = seq_heap
            else:
                backward_seq = (i for i in seq_heap if i < max_item)

            for i in backward_seq:
                i.backward()

        except StopIteration:
            break



import os
import sys


class ChildForkFinished(Exception):
    pass


class DeamonProcess(object):

    def __init__(
        self, target=None, pidfile=None,
        stdin=None, stdout=None, stderr=None,
    ):
        self.target = target
        self._pid = None
        self.pidfile = pidfile or "/tmp/DeamonProcess.pid"
        self.stdin = stdin or "/dev/null"
        self.stdout = stdout or "/dev/null"
        self.stderr = stderr or self.stdout

    @property
    def pid(self):
        if self._pid is None:
            try:
                with open(self.pidfile, "rt") as fp:
                    self._pid = int(fp.readline())
            except (IOError, ValueError):
                pass
        return self._pid

    def run(self, *args, **kwargs):
        if not self.target:
            raise NotImplementedError()
        self.target(*args, **kwargs)

    def _fork_child_process(self):
        """
        Fork a child process which would not collected by the init.
        """
        try:
            pid = os.fork()
        except OSError as err:
            sys.stderr.write("fork #1 failed: %s" % err)

        if pid > 0:
            raise ChildForkFinished()

    def _wait_child_process(self):
        """
        Wait child process to exit to avoid a zombie process.
        """
        return os.wait()

    def _fork_deamon_process(self):
        """
        Grandchild process adopted by the init because child process
        has been exited.
        """
        try:
            pid = os.fork()
        except OSError as err:
            sys.stderr.write("fork #1 failed: %s" % err)
            sys.exit(1)

        if pid > 0:
            sys.exit(0)

    def _redirect_standard_file_descriptors(self):
        """
        Redirect stdin, stdout, stderr.
        """
        sys.stderr.flush()

        stdin = open(self.stdin, "r")
        os.dup2(stdin.fileno(), sys.stdin.fileno())

        stdout = open(self.stdout, "a+")
        sys.stdout.flush()
        os.dup2(stdout.fileno(), sys.stdout.fileno())

        stderr = open(self.stderr, "a+")
        sys.stderr.flush()
        os.dup2(stderr.fileno(), sys.stderr.fileno())

    def _daemonize(self):
        # move to background
        try:
            self._fork_child_process()
        except ChildForkFinished as err:
            self._wait_child_process()
            raise err

        # become session leader
        os.setsid()
        # change directory to root to avoid unmount error
        os.chdir("/")
        # reset the default umask file creation mask
        os.umask(0)

        # become process group leader
        self._fork_deamon_process()
        with open(self.pidfile, "w+") as fp:
            self._pid = os.getpid()
            fp.write("%s\n" % self._pid)

        # redirect standard file descriptors
        self._redirect_standard_file_descriptors()

    def start(self, args=(), kwargs=None):
        kwargs = kwargs or {}
        try:
            self._daemonize()
        except ChildForkFinished:
            return

        self.run(*args, **kwargs)



#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase

import mock

from ycyc.tests import mock_patches
from ycyc.base import filetools


class TestOserrorFormat(TestCase):
    def test_windows_env(self):
        with mock_patches("ycyc.base.filetools.exceptions") as patches:
            patches.exceptions.WindowsError = NotImplementedError

            @filetools.oserror_format
            def foo():
                raise patches.exceptions.WindowsError

            with self.assertRaises(OSError):
                foo()

    def test_linux_env(self):
        @filetools.oserror_format
        def foo():
            raise OSError

        with self.assertRaises(OSError):
            foo()


class TestDirHelper(TestCase):
    Path = "./lyc/"

    def test_make_sure_dir_exists(self):
        with mock_patches(
            "ycyc.base.filetools.os.path.exists",
            "ycyc.base.filetools.os.makedirs",
        ) as patches:
            patches.exists.return_value = True

            filetools.make_sure_dir_exists(self.Path)

            self.assertEqual(patches.exists.call_count, 1)
            self.assertEqual(patches.makedirs.call_count, 0)

        with mock_patches(
            "ycyc.base.filetools.os.path.exists",
            "ycyc.base.filetools.os.makedirs",
        ) as patches:
            patches.exists.return_value = False

            filetools.make_sure_dir_exists(self.Path)

            self.assertEqual(patches.exists.call_count, 1)
            self.assertEqual(patches.makedirs.call_count, 1)

    def test_remove_dir(self):
        with mock_patches(
            "ycyc.base.filetools.shutil.rmtree",
            "ycyc.base.filetools.os.removedirs",
        ) as patches:
            filetools.remove_dir(self.Path)

            self.assertEqual(patches.rmtree.call_count, 0)
            self.assertEqual(patches.removedirs.call_count, 1)

        with mock_patches(
            "ycyc.base.filetools.shutil.rmtree",
            "ycyc.base.filetools.os.removedirs",
        ) as patches:
            filetools.remove_dir(self.Path, True)

            self.assertEqual(patches.rmtree.call_count, 1)
            self.assertEqual(patches.removedirs.call_count, 0)

    def test_make_sure_dir_empty(self):
        with mock_patches(
            "ycyc.base.filetools.os.path.exists",
            "ycyc.base.filetools.make_sure_dir_exists",
        ) as patches:
            patches.exists.return_value = False

            filetools.make_sure_dir_empty(self.Path)

            patches.exists.assert_called_once_with(self.Path)
            patches.make_sure_dir_exists.assert_called_once_with(self.Path)

        with mock_patches(
            "ycyc.base.filetools.os.path.exists",
            "ycyc.base.filetools.os.listdir",
            "ycyc.base.filetools.os.path.isdir",
            "ycyc.base.filetools.remove_dir",
            "ycyc.base.filetools.os.remove",
        ) as patches:
            patches.exists.return_value = True
            patches.listdir.return_value = ["a", "b/"]
            patches.isdir.side_effect = lambda f: f.endswith("/")

            filetools.make_sure_dir_empty(self.Path)

            patches.exists.assert_called_once_with(self.Path)
            patches.listdir.assert_called_once_with(self.Path)
            patches.remove.assert_called_once_with(self.Path + "a")
            patches.remove_dir.assert_called_once_with(self.Path + "b/")

    def test_make_sure_not_exists(self):
        with mock_patches(
            "ycyc.base.filetools.os.path.exists",
        ) as patches:
            patches.exists.return_value = False

            filetools.make_sure_not_exists(self.Path)

            patches.exists.assert_called_once_with(self.Path)

        with mock_patches(
            "ycyc.base.filetools.os.path.exists",
            "ycyc.base.filetools.os.path.isdir",
            "ycyc.base.filetools.remove_dir",
        ) as patches:
            patches.exists.return_value = True
            patches.isdir.return_value = True

            filetools.make_sure_not_exists(self.Path)

            patches.exists.assert_called_once_with(self.Path)
            patches.isdir.assert_called_once_with(self.Path)
            patches.remove_dir.assert_called_once_with(self.Path, recursion=True)

        with mock_patches(
            "ycyc.base.filetools.os.path.exists",
            "ycyc.base.filetools.os.path.isdir",
            "ycyc.base.filetools.os.remove",
        ) as patches:
            patches.exists.return_value = True
            patches.isdir.return_value = False

            filetools.make_sure_not_exists(self.Path)

            patches.exists.assert_called_once_with(self.Path)
            patches.isdir.assert_called_once_with(self.Path)
            patches.remove.assert_called_once_with(self.Path)


class TestChangeDir(TestCase):
    Path = "./lyc/"

    def test_usage(self):
        with mock_patches(
            "ycyc.base.filetools.os.chdir",
        ) as patches:
            with filetools.cd(self.Path):
                self.assertEqual(patches.chdir.call_count, 1)

            self.assertEqual(patches.chdir.call_count, 2)

    def test_exception(self):
        with mock_patches(
            "ycyc.base.filetools.os.chdir",
        ) as patches:
            with self.assertRaises(NotImplementedError):
                with filetools.cd(self.Path):
                    self.assertEqual(patches.chdir.call_count, 1)
                    raise NotImplementedError

            self.assertEqual(patches.chdir.call_count, 2)



from ycyc.ycollections.heap import Heap


class SerialListItem(object):
    def __init__(self, sn, value):
        self.sn = sn
        self.value = value


class SerialList(object):
    def __init__(self, sn=0, reverse=False):
        self.next_sn = sn
        self.heap = Heap(cmp_attrs=["sn"], reverse=False)

    def push_item(self, item):
        self.heap.push(item)

    def push(self, sn, value):
        self.push_item(SerialListItem(sn, value))

    def pop_item(self, force=False):
        while True:
            items = self.heap.headn(1)
            if not items:
                break
            item = items[0]

            if item.sn < self.next_sn:
                self.heap.pop()
            elif item.sn == self.next_sn or force:
                item = self.heap.pop()
                self.next_sn = item.sn + 1
                return item
            else:
                break
        return None

    def pop(self, force=False):
        item = self.pop_item(force)
        return item and item.value



#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase
from datetime import datetime, timedelta

from ycyc.ycollections import rangecls


class TestARange(TestCase):
    def test_constructor(self):
        r = rangecls.ARange(0, 2)
        self.assertTrue(r.include_start)
        self.assertTrue(r.include_end)

        with self.assertRaises(rangecls.RangeError):
            rangecls.ARange(5, 2)

    def test_contains(self):
        r = rangecls.ARange(0, 2)
        self.assertTrue(0 in r)
        self.assertTrue(1 in r)
        self.assertTrue(2 in r)

        r = rangecls.ARange(0, 2, include_start=False, include_end=False)
        self.assertFalse(0 in r)
        self.assertTrue(1 in r)
        self.assertFalse(2 in r)

    def test_overlapping(self):
        r1 = rangecls.ARange(0, 5)
        r2 = rangecls.ARange(1, 3)

        r3 = r1 & r2
        self.assertEqual(r3.start, 1)
        self.assertEqual(r3.end, 3)
        self.assertEqual(r3.include_start, True)
        self.assertEqual(r3.include_end, True)

        r4 = r2 & r1
        self.assertEqual(r4.start, 1)
        self.assertEqual(r4.end, 3)
        self.assertEqual(r4.include_start, True)
        self.assertEqual(r4.include_end, True)

        r1 = rangecls.ARange(0, 2)
        r2 = rangecls.ARange(1, 3)

        r3 = r1 & r2
        self.assertEqual(r3.start, 1)
        self.assertEqual(r3.end, 2)
        self.assertEqual(r3.include_start, True)
        self.assertEqual(r3.include_end, True)

        r4 = r2 & r1
        self.assertEqual(r4.start, 1)
        self.assertEqual(r4.end, 2)
        self.assertEqual(r4.include_start, True)
        self.assertEqual(r4.include_end, True)

        r1 = rangecls.ARange(1, 2)
        r2 = rangecls.ARange(4, 5)
        self.assertEqual(r1 & r2, None)
        self.assertEqual(r2 & r1, None)

    def test_point_position(self):
        r = rangecls.ARange(1, 3)
        self.assertTrue(r.is_after_of(0))
        self.assertTrue(r.is_before_of(4))
        self.assertFalse(r.is_after_of(2))
        self.assertFalse(r.is_before_of(2))
        self.assertFalse(r.is_after_of(4))
        self.assertFalse(r.is_before_of(0))

        r = rangecls.ARange(0, 2, include_start=False, include_end=False)
        self.assertFalse(r.is_after_of(2))
        self.assertTrue(r.is_before_of(2))
        self.assertTrue(r.is_after_of(0))
        self.assertFalse(r.is_before_of(0))

    def test_sub_range(self):
        r1 = rangecls.ARange(0, 5)
        r2 = rangecls.ARange(1, 3)
        self.assertTrue(r1.has_sub_range(r2))
        self.assertTrue(r1.has_sub_range(r1))
        self.assertFalse(r2.has_sub_range(r1))
        self.assertTrue(r2.has_sub_range(r2))

        r3 = rangecls.ARange(0, 2, include_start=False, include_end=False)
        r4 = rangecls.ARange(0, 2)
        r5 = rangecls.ARange(0, 3, include_start=False)
        self.assertTrue(r4.has_sub_range(r3))
        self.assertTrue(r5.has_sub_range(r3))
        self.assertFalse(r5.has_sub_range(r4))

    def test_is_follow_by(self):
        r1 = rangecls.ARange(0, 3)
        r2 = rangecls.ARange(0, 3, include_end=False)
        r3 = rangecls.ARange(3, 6)
        r4 = rangecls.ARange(3, 6, include_start=False)
        r5 = rangecls.ARange(4, 6)

        self.assertFalse(r1.is_follow_by(r3))
        self.assertTrue(r1.is_follow_by(r4))
        self.assertFalse(r1.is_follow_by(r5))

        self.assertTrue(r2.is_follow_by(r3))
        self.assertFalse(r2.is_follow_by(r4))
        self.assertFalse(r2.is_follow_by(r5))

    def test_sampling(self):
        r = rangecls.ARange(0, 4)
        self.assertListEqual(list(r.sampling(1)), [0, 1, 2, 3, 4])
        self.assertListEqual(list(r.sampling(2)), [0, 2, 4])

        r = rangecls.ARange(0, 4, include_start=False)
        self.assertListEqual(list(r.sampling(2)), [2, 4])

        r = rangecls.ARange(0, 4, include_start=False, include_end=False)
        self.assertListEqual(list(r.sampling(2)), [2])
        self.assertListEqual(list(r.sampling(1.25)), [1.25, 2.5, 3.75])

        r = rangecls.ARange(datetime(2010, 12, 18), datetime(2015, 4, 26))
        self.assertEqual(len(list(r.sampling(timedelta(days=1)))), 1591)



#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase

from ycyc.base import txtutils


class TestCoding(TestCase):
    b_en = b"test"
    b_ch = u"测试".encode("utf-8")
    u_en = u"test"
    u_ch = u"测试"

    def test_encode(self):
        self.assertEqual(self.b_en, txtutils.encode(self.u_en, "utf-8"))
        self.assertEqual(self.b_ch, txtutils.encode(self.u_ch, "utf-8"))
        self.assertEqual(self.b_en, txtutils.encode(self.b_en, "utf-8"))
        self.assertEqual(self.b_ch, txtutils.encode(self.b_ch, "utf-8"))

    def test_decode(self):
        self.assertEqual(self.u_en, txtutils.decode(self.u_en, "utf-8"))
        self.assertEqual(self.u_ch, txtutils.decode(self.u_ch, "utf-8"))
        self.assertEqual(self.u_en, txtutils.decode(self.b_en, "utf-8"))
        self.assertEqual(self.u_ch, txtutils.decode(self.b_ch, "utf-8"))

    def test_bytescat(self):
        target = u"test测试".encode("utf-8")
        self.assertEqual(txtutils.bytescat(self.b_en, self.b_ch), target)
        self.assertEqual(txtutils.bytescat(self.u_en, self.u_ch), target)
        self.assertEqual(txtutils.bytescat(self.b_en, self.u_ch), target)
        self.assertEqual(txtutils.bytescat(self.u_en, self.b_ch), target)

        target = u"测试测试".encode("utf-8")
        self.assertEqual(txtutils.bytescat(self.b_ch, self.b_ch), target)
        self.assertEqual(txtutils.bytescat(self.u_ch, self.u_ch), target)
        self.assertEqual(txtutils.bytescat(self.b_ch, self.u_ch), target)
        self.assertEqual(txtutils.bytescat(self.u_ch, self.b_ch), target)

    def test_strcat(self):
        target = u"test测试"
        self.assertEqual(txtutils.strcat(self.b_en, self.b_ch), target)
        self.assertEqual(txtutils.strcat(self.u_en, self.u_ch), target)
        self.assertEqual(txtutils.strcat(self.b_en, self.u_ch), target)
        self.assertEqual(txtutils.strcat(self.u_en, self.b_ch), target)

        target = u"测试测试"
        self.assertEqual(txtutils.strcat(self.b_ch, self.b_ch), target)
        self.assertEqual(txtutils.strcat(self.u_ch, self.u_ch), target)
        self.assertEqual(txtutils.strcat(self.b_ch, self.u_ch), target)
        self.assertEqual(txtutils.strcat(self.u_ch, self.b_ch), target)

        gbk_ch = u"测试".encode("gbk")
        self.assertEqual(
            txtutils.strcat(gbk_ch, self.b_ch, "gbk", "utf-8"), target)
        self.assertEqual(
            txtutils.strcat(gbk_ch, self.u_ch, "gbk", "utf-8"), target)
        self.assertEqual(
            txtutils.strcat(self.b_ch, gbk_ch, "utf-8", "gbk"), target)
        self.assertEqual(
            txtutils.strcat(self.u_ch, gbk_ch, "utf-8", "gbk"), target)


class TestStringUtils(TestCase):
    def test_drop_prefix(self):
        self.assertEqual(
            txtutils.drop_prefix("root/file.txt", "root/"),
            "file.txt"
        )
        self.assertEqual(
            txtutils.drop_prefix("dir/file.txt", "root/"),
            "dir/file.txt"
        )

    def test_drop_postfix(self):
        self.assertEqual(
            txtutils.drop_postfix("file.txt", ".txt"),
            "file"
        )
        self.assertEqual(
            txtutils.drop_postfix("file.tmp", ".txt"),
            "file.tmp"
        )

    def test_split_and_strip(self):
        self.assertListEqual(
            txtutils.split_and_strip("a,b ,c, d , e"),
            ["a", "b", "c", "d", "e"]
        )

    def test_left_part_of(self):
        txt = "xxx.avi.part1.rar"
        self.assertEqual(txtutils.left_part_of(txt, "."), "xxx")
        self.assertEqual(txtutils.left_part_of(txt, ".", 0), "")
        self.assertEqual(txtutils.left_part_of(txt, ".", 1), "xxx")
        self.assertEqual(txtutils.left_part_of(txt, ".", 2), "xxx.avi")
        self.assertEqual(txtutils.left_part_of(txt, ".", -1), "xxx.avi.part1")
        self.assertEqual(txtutils.left_part_of(txt, ".", -2), "xxx.avi")

    def test_right_part_of(self):
        txt = "xxx.avi.part1.rar"
        self.assertEqual(txtutils.right_part_of(txt, "."), "rar")
        self.assertEqual(txtutils.right_part_of(txt, ".", 0), txt)
        self.assertEqual(txtutils.right_part_of(txt, ".", -1), "rar")
        self.assertEqual(txtutils.right_part_of(txt, ".", -2), "part1.rar")
        self.assertEqual(txtutils.right_part_of(txt, ".", 1), "avi.part1.rar")
        self.assertEqual(txtutils.right_part_of(txt, ".", 2), "part1.rar")


class TestTxtDistance(TestCase):
    def test_edit_distance(self):
        test_cases = {
            ("ABC", "CBCD"): 2,
            ("FOOD", "MONEY"): 4,
            ("ISLANDER", "SLANDER"): 1,
            ("MART", "KARMA"): 3,
            ("SITTING", "SITTING"): 0,
            ("INTENTION", "EXECUTION"): 5,
            ("", "AA"): 2,
            ("B", ""): 1,
        }
        edit_distance = txtutils.TxtDistance.edit_distance
        edit_distance2 = txtutils.TxtDistance.edit_distance2

        for case, result in test_cases.items():
            s1, s2 = case
            self.assertEqual(edit_distance(s1, s2), result)
            self.assertEqual(edit_distance2(s1, s2), result)

    def test_hamming_distance(self):
        test_cases = {
            ("12345", "54321"): 4,
            ("karolin", "kathrin"): 3,
            ("karolin", "kerstin"): 3,
            ("1011101", "1001001 "): 2,
            ("2173896", "2233796"): 3,
        }

        for case, result in test_cases.items():
            s1, s2 = case
            self.assertEqual(
                txtutils.TxtDistance.hamming_distance(s1, s2),
                result,
            )


class TestSepJoin(TestCase):
    def test_usage(self):
        self.assertEqual(
            txtutils.sep_join("|", map(str, range(3))),
            "0|1|2"
        )
        self.assertEqual(
            txtutils.sep_join("|", map(str, range(3)), True),
            "|0|1|2"
        )
        self.assertEqual(
            txtutils.sep_join("|", map(str, range(3)), True, True),
            "|0|1|2|"
        )
        self.assertEqual(
            txtutils.sep_join(" | ", map(str, range(3)), "| ", " |"),
            "| 0 | 1 | 2 |"
        )
        self.assertEqual(
            txtutils.sep_join(",", map(str, range(3)), "x="),
            "x=0,1,2"
        )



#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase
import sys
import thread

from ycyc.base import logutils


class TestLoggerInfo(TestCase):
    def test_usage(self):
        frames = sys._current_frames()
        frame = frames[thread.get_ident()]
        loginfo = logutils.LoggerInfo()

        self.assertIs(frame, loginfo.frame)
        line_based = 18
        self.assertEqual(line_based + 1, loginfo.line_no)
        self.assertEqual(line_based + 2, loginfo.line_no)
        self_func = self.test_usage.im_func
        self.assertEqual(self_func.__name__, loginfo.code_name)



#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase

from ycyc.contribs.tail_call_optimized import tail_call_optimized


class TestTailCallOptimized(TestCase):
    def test_usage(self):
        @tail_call_optimized
        def factorial(n, acc=1):
            "calculate a factorial"
            if n == 0:
                return acc
            return factorial(n - 1, n * acc)

        factorial(10000)

        @tail_call_optimized
        def fib(i, current=0, next=1):
            if i == 0:
                return current
            else:
                return fib(i - 1, next, current + next)

        fib(10000)



#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase

from ycyc.base import typeutils


class TestSubType(TestCase):
    def test_usage(self):
        self.assertTrue(issubclass(
            typeutils.subtype("Test"),
            object
        ))
        self.assertTrue(issubclass(
            typeutils.subtype("Test", dict),
            dict
        ))
        self.assertTrue(issubclass(
            typeutils.subtype("Test", (dict,)),
            dict
        ))
        TestType = typeutils.subtype("Test", object, {"val": 1})
        self.assertEqual(TestType.__name__, "Test")
        self.assertEqual(TestType.val, 1)
        obj = TestType()
        self.assertEqual(obj.val, 1)


class TestTypesFactory(TestCase):
    def test_usage(self):
        TestType = typeutils.types_factory.TestType()
        self.assertTrue(issubclass(TestType, object))
        self.assertIsNot(TestType, typeutils.types_factory.TestType())
        self.assertTrue(issubclass(
            typeutils.types_factory.TestType(dict),
            dict
        ))
        TestType = typeutils.types_factory.TestType(attrs={"val": 1})
        self.assertEqual(TestType.__name__, "TestType")
        self.assertEqual(TestType.val, 1)
        obj = TestType()
        self.assertEqual(obj.val, 1)


class TestSubException(TestCase):
    def test_usage(self):
        self.assertTrue(issubclass(
            typeutils.subexception("Test"),
            Exception
        ))


class TestExceptionsFactory(TestCase):
    def test_usage(self):
        TestType = typeutils.exceptions_factory.TestType()
        self.assertTrue(issubclass(TestType, Exception))
        self.assertIsNot(TestType, typeutils.exceptions_factory.TestType())


class TestSimpleExceptions(TestCase):
    def test_usage(self):
        simple_exceptions = typeutils.SimpleExceptions()
        self.assertIsInstance(simple_exceptions.NewError, type)
        self.assertIs(simple_exceptions.NewError, simple_exceptions.NewError)


class TestFreezedAttrs(TestCase):
    def test_usage(self):
        @typeutils.freezed_attrs(["name", "value"])
        class TestObj(object):
            def __init__(self, name, value):
                self.name = name
                self.value = value

        m = TestObj("test", 123)
        with self.assertRaisesRegexp(AttributeError, "name is not writable"):
            m.name = 1
        self.assertEqual(m.name, "test")
        with self.assertRaisesRegexp(AttributeError, "value is not writable"):
            m.value = 1
        self.assertEqual(m.value, 123)
        with self.assertRaisesRegexp(AttributeError, "noting is not writable"):
            m.noting = 1
        self.assertFalse(hasattr(m, "noting"))


class TestConstants(TestCase):
    def test_usage(self):
        const = typeutils.constants(
            Id="TestConstants",
            Name="TestConstants",
            Function="test_usage",
        )

        self.assertIsInstance(const, typeutils.Constants)
        self.assertEqual(const.Id, "TestConstants")  # pylint: disable=E1101
        self.assertEqual(const.Name, "TestConstants")  # pylint: disable=E1101
        self.assertTupleEqual(const["TestConstants"], ("Id", "Name"))
        self.assertEqual(const.Function, "test_usage")  # pylint: disable=E1101
        self.assertTupleEqual(const["test_usage"], ("Function",))
        self.assertIsNone(const["noting"])

        with self.assertRaisesRegexp(
            AttributeError,
            "attribute Name is not writable",
        ):
            const.Name = "error"

        with self.assertRaisesRegexp(
            AttributeError,
            "attribute Function is not writable",
        ):
            const.Function = "error"

        self.assertDictEqual(dict(const), {
            "Id": "TestConstants",
            "Name": "TestConstants",
            "Function": "test_usage",
        })


class TestEnums(TestCase):
    def test_usage(self):
        enums = typeutils.enums("a", "b", "c")

        self.assertEqual(enums.a, 0)  # pylint: disable=E1101
        self.assertEqual(enums.b, 1)  # pylint: disable=E1101
        self.assertEqual(enums.c, 2)  # pylint: disable=E1101

        self.assertEqual(enums[0], "a")
        self.assertEqual(enums[1], "b")
        self.assertEqual(enums[2], "c")

        with self.assertRaisesRegexp(
            AttributeError,
            "attribute a is not writable",
        ):
            enums.a = 2

        self.assertDictEqual(dict(enums), {
            "a": 0, "b": 1, "c": 2,
        })



#!/usr/bin/env python
# encoding: utf-8

from collections import namedtuple

from ycyc.base.contextutils import catch
from ycyc.base.typeutils import SimpleExceptions

EventsExceptions = SimpleExceptions()
ListenerDuplicatedError = EventsExceptions.ListenerDuplicatedError
ListenerNoExistedError = EventsExceptions.ListenerNoExistedError
EventNoExistedError = EventsExceptions.EventNoExistedError


class Event(object):
    CallbackResult = namedtuple("CallbackResult", [
        "callback", "result", "exception",
    ])

    def __init__(self):
        super(Event, self).__init__()
        self.callbacks = []

    def register(self, callback):
        """
        Register a callback to this event
        """
        if callback in self.callbacks:
            raise ListenerDuplicatedError()
        self.callbacks.append(callback)
        return callback

    def unregister(self, callback):
        """
        Unregister a callback from this event
        """
        if callback not in self.callbacks:
            raise ListenerNoExistedError()
        self.callbacks.remove(callback)
        return callback

    def notify_iter(self, sender, args=None, kwargs=None):
        args = args or tuple()
        kwargs = kwargs or dict()

        for callback in self.callbacks:
            try:
                exception = None
                result = callback(sender, *args, **kwargs)
            except Exception as exception:
                result = None

            yield self.CallbackResult(
                callback=callback, result=result,
                exception=exception,
            )

    def notify(self, sender, *args, **kwargs):
        """
        Notify the listeners when event is happending,
        but stop if catch a exception.
        """
        for i in self.notify_iter(sender, *args, **kwargs):
            if i.exception:
                raise i.exception

    def notify_all(self, sender, *args, **kwargs):
        """
        Notify all the listeners when event is happending.
        """
        return [
            i for i in self.notify_iter(sender, *args, **kwargs)
        ]


class EventsCenter(dict):
    """
    Event center to manage some events as a dict
    """
    def __init__(self, event_factory=None):
        super(EventsCenter, self).__init__()
        self.event_factory = event_factory or Event

    def __missing__(self, name):
        event = self[name] = self.event_factory()
        return event



#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase

from ycyc.ycollections import table


class TestSimpleTable(TestCase):
    def test_usage(self):
        tb = table.simple_table(3, 4, lambda x, y: x * 10 + y)
        self.assertListEqual(tb[0], [0, 1, 2, 3])
        self.assertListEqual(tb[1], [10, 11, 12, 13])
        self.assertListEqual(tb[2], [20, 21, 22, 23])


class TestTable(TestCase):
    def test_usage(self):
        tb = table.Table(3, 4, 0)
        self.assertEqual(len(tb.table), 3)
        self.assertEqual(len(tb.table[0]), 4)

        self.assertEqual(tb[1, 2], 0)
        tb[1, 2] = 1
        self.assertEqual(tb[1, 2], 1)


class TestSparseTable(TestCase):
    def test_usage(self):
        tb = table.SparseTable(3, 4, 0)
        self.assertEqual(tb[1, 2], 0)
        tb[1, 2] = 1
        self.assertEqual(tb[1, 2], 1)



#!/usr/bin/env python
# encoding: utf-8

import re
from unittest import TestCase

from ycyc.base import resources


class TestRegex(TestCase):
    def pattern_equal_rex(self, pattern):
        return re.compile(pattern.rstrip("$") + "$")

    def test_num_less_than(self):
        with self.assertRaises(ValueError):
            rex = re.compile(resources.Regex.num_less_than(0))

        def test_num(num):
            rex = self.pattern_equal_rex(resources.Regex.num_less_than(num))
            for i in range(num + num / 2):
                if i < num:
                    self.assertIsNotNone(rex.match(str(i)))
                else:
                    self.assertIsNone(rex.match(str(i)))

        test_num(1)
        test_num(2)

        test_num(9)
        test_num(10)
        test_num(11)
        test_num(12)

        test_num(99)
        test_num(100)
        test_num(101)
        test_num(102)

        test_num(200)
        test_num(201)
        test_num(255)
        test_num(256)

        test_num(999)
        test_num(1000)
        test_num(1001)
        test_num(1010)
        test_num(1100)

        test_num(1991)
        test_num(1999)
        test_num(2000)
        test_num(2001)
        test_num(2002)



#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase
import mock
import time

from ycyc.ycollections.namedict import namedict, RequireFieldsMissError


class TestNamedDict(TestCase):
    def test_usage(self):
        cur_time = time.time()
        Params = namedict(
            "InvokeParams",
            requires=("sender", "callback"),
            fields={
                "value": None,
                "time": cur_time,
            },
        )

        self.assertEqual(Params.__name__, "InvokeParams")

        with self.assertRaisesRegexp(
            RequireFieldsMissError, ", ".join({"sender", "callback"})
        ):
            Params()

        with self.assertRaisesRegexp(RequireFieldsMissError, "sender"):
            Params(callback=None)

        with self.assertRaisesRegexp(RequireFieldsMissError, "callback"):
            Params(sender=None)

        params = Params("test", 1)
        self.assertIsInstance(params, dict)
        self.assertEqual(params.sender, "test")
        self.assertEqual(params.callback, 1)
        self.assertEqual(params.value, None)
        self.assertEqual(params.time, cur_time)

        params = Params(sender="test", callback=2)
        self.assertIsInstance(params, dict)
        self.assertEqual(params.sender, "test")
        self.assertEqual(params.callback, 2)
        self.assertEqual(params.value, None)
        self.assertEqual(params.time, cur_time)

        params = Params("test", None, value=1)
        self.assertIsInstance(params, dict)
        self.assertEqual(params.sender, "test")
        self.assertEqual(params.callback, None)
        self.assertEqual(params.value, 1)
        self.assertEqual(params.time, cur_time)

        params = Params("test", None, value=1, time=2)
        self.assertIsInstance(params, dict)
        self.assertEqual(params.sender, "test")
        self.assertEqual(params.callback, None)
        self.assertEqual(params.value, 1)
        self.assertEqual(params.time, 2)

        Params = namedict(
            "InvokeParams",
            requires=("sender", "callback"),
            fields=(
                ("value", None),
                ("time", cur_time),
            ),
        )

        params = Params("test", None, 1)
        self.assertIsInstance(params, dict)
        self.assertEqual(params.sender, "test")
        self.assertEqual(params.callback, None)
        self.assertEqual(params.value, 1)
        self.assertEqual(params.time, cur_time)

        params = Params("test", None, 1, 2)
        self.assertIsInstance(params, dict)
        self.assertEqual(params.sender, "test")
        self.assertEqual(params.callback, None)
        self.assertEqual(params.value, 1)
        self.assertEqual(params.time, 2)

        params = Params("test", None, 1, value=2)
        self.assertIsInstance(params, dict)
        self.assertEqual(params.sender, "test")
        self.assertEqual(params.callback, None)
        self.assertEqual(params.value, 1)
        self.assertEqual(params.time, cur_time)

        params = Params("test", None, 1, value=2, nothing=3)
        self.assertIsInstance(params, dict)
        self.assertEqual(params.sender, "test")
        self.assertEqual(params.callback, None)
        self.assertEqual(params.value, 1)
        self.assertEqual(params.time, cur_time)
        with self.assertRaises(AttributeError):
            params.nothing

        Params = namedict("Params", ["key", "value"])
        params = Params("test", value=2)
        self.assertEqual(params.key, "test")
        self.assertEqual(params.value, 2)

    def test_logging(self):
        with mock.patch("ycyc.ycollections.namedict.logger"):
            from ycyc.ycollections.namedict import logger
            TestDict = namedict("TestDict", {"val": None})

            tdict = TestDict(1)
            self.assertEqual(tdict.val, 1)
            self.assertEqual(logger.warning.call_count, 0)

            tdict = TestDict(1, 2)
            self.assertEqual(tdict.val, 1)
            self.assertEqual(logger.warning.call_count, 1)
            self.assertListEqual(logger.warning.call_args[0][1], [2])

            tdict = TestDict(1, nothing=2)
            self.assertEqual(tdict.val, 1)
            self.assertEqual(logger.warning.call_count, 2)
            self.assertSetEqual(logger.warning.call_args[0][1], {"nothing"})



#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase

from ycyc.base import marker


class TestMarker(TestCase):
    def test_usage(self):
        self.assertNotEqual(marker.Marker(), marker.Marker())

        test1 = marker.Marker("test")
        test2 = marker.Marker("test")
        self.assertEqual(test1, test2)
        self.assertIsNot(test1, test2)

        test_undefined = marker.Marker("undefined")
        self.assertEqual(test_undefined, marker.Marker.Undefined)
        self.assertIsNot(test_undefined, marker.Marker.Undefined)

        m = marker.Marker()
        self.assertEqual(m, m)
        self.assertIs(m, m)

        self.assertFalse(bool(marker.Marker.Undefined))
        self.assertFalse(bool(marker.Marker.Missed))
        self.assertFalse(bool(marker.Marker.Disabled))

        m = marker.Marker("test", True)
        self.assertTrue(m)

    def test_freeze_attrs(self):
        m = marker.Marker("test", 123)
        with self.assertRaisesRegexp(AttributeError, "name is not writable"):
            m.name = 1
        self.assertEqual(m.name, "test")
        with self.assertRaisesRegexp(AttributeError, "value is not writable"):
            m.value = 1
        self.assertEqual(m.value, 123)
        with self.assertRaisesRegexp(AttributeError, "has no attribute 'foo'"):
            m.foo = 1  # pylint: disable=E0237
        self.assertFalse(hasattr(m, "foo"))



from unittest import TestCase

from ycyc.ycollections.heap import Heap, HeapIsEmpty


class TestHeap(TestCase):
    class PriorityTask(object):
        def __init__(self, priority, value):
            self.priority = priority
            self.value = value

    def test_construction(self):
        Heap()
        Heap([2, 3, 1, 17, 19, 100, 25, 17, 36])
        Heap(range(10), cmp_attrs=["real"])
        Heap(range(10), reverse=True)

    def test_iter(self):
        heap = Heap([2, 3, 1, 17, 19, 100, 25, 17, 36])
        self.assertListEqual(
            list(heap),
            [1, 2, 3, 17, 17, 19, 25, 36, 100]
        )

        heap = Heap([2, 3, 1, 17, 19, 100, 25, 17, 36], reverse=True)
        self.assertListEqual(
            list(heap),
            [100, 36, 25, 19, 17, 17, 3, 2, 1]
        )

        heap = Heap([2, 3, 1, 17, 19, 100, 25, 17, 36], cmp_attrs=["real"])
        self.assertListEqual(
            list(heap),
            [1, 2, 3, 17, 17, 19, 25, 36, 100]
        )

    def test_push(self):
        heap = Heap()
        heap.push(3)
        self.assertListEqual(list(heap), [3])
        heap.push(1)
        self.assertListEqual(list(heap), [1, 3])
        heap.push(2)
        self.assertListEqual(list(heap), [1, 2, 3])

    def test_pop(self):
        heap = Heap([3, 2, 1])
        self.assertListEqual(list(heap), [1, 2, 3])
        heap.pop()
        self.assertListEqual(list(heap), [2, 3])
        heap.pop()
        self.assertListEqual(list(heap), [3])
        heap.pop()
        self.assertListEqual(list(heap), [])
        with self.assertRaises(HeapIsEmpty):
            heap.pop()

    def test_headn(self):
        heap = Heap([3, 1, 17, 25, 19])
        self.assertListEqual(
            heap.headn(3),
            [1, 3, 17]
        )
        heap = Heap([3, 1, 17, 25, 19], reverse=True)
        self.assertListEqual(
            heap.headn(3),
            [25, 19, 17]
        )

    def test_tailn(self):
        heap = Heap([3, 1, 17, 25, 19])
        self.assertListEqual(
            heap.tailn(3),
            [17, 19, 25]
        )

        heap = Heap([3, 1, 17, 25, 19], reverse=True)
        self.assertListEqual(
            heap.tailn(3),
            [17, 3, 1]
        )

    def test_edge_out(self):
        heap = Heap([3, 1, 2])
        self.assertEqual(1, heap.edge_out(4))
        self.assertEqual(2, heap.edge_out(5))
        self.assertEqual(3, heap.edge_out(6))
        self.assertEqual(4, heap.edge_out(7))

    def test_cmp_attrs(self):
        heap = Heap(cmp_attrs=["priority"])
        heap.push(self.PriorityTask(5, 0))
        heap.push(self.PriorityTask(8, 1))
        heap.push(self.PriorityTask(3, 2))
        heap.push(self.PriorityTask(4, 3))
        heap.push(self.PriorityTask(9, 4))
        self.assertListEqual(
            [i.value for i in heap],
            [2, 3, 0, 1, 4]
        )

        heap = Heap(cmp_attrs=["priority"], reverse=True)
        heap.push(self.PriorityTask(5, 0))
        heap.push(self.PriorityTask(8, 1))
        heap.push(self.PriorityTask(3, 2))
        heap.push(self.PriorityTask(4, 3))
        heap.push(self.PriorityTask(9, 4))
        self.assertListEqual(
            [i.value for i in heap],
            [4, 1, 0, 3, 2]
        )



#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase

from ycyc.ycollections import filterlist


class TestFilterList(TestCase):
    def test_usage(self):
        list1 = filterlist.FilterList(range(5))
        list2 = list1.filter(lambda x: x % 2)
        list3 = list1.exclude(lambda x: x % 2)

        self.assertIsInstance(list2, filterlist.FilterList)
        self.assertIsInstance(list3, filterlist.FilterList)
        self.assertEqual(list(list2), [1, 3])
        self.assertEqual(list(list3), [0, 2, 4])
        self.assertEqual(list1.first(), 0)
        self.assertEqual(list1.last(), 4)

        list1.clear()
        self.assertEqual(list1.first(), None)
        self.assertEqual(list1.last(), None)
        self.assertEqual(list1.first(0), 0)
        self.assertEqual(list1.last(0), 0)



#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase

import mock

from ycyc.tests import mock_patches
from ycyc.base import shelltools


class TestCommands(TestCase):
    def test_shell_commands_usage(self):
        self.assertIsInstance(shelltools.ShellCommands, object)
        command = shelltools.ShellCommands.echo  # pylint: disable=E1101
        self.assertIsInstance(command, shelltools.Command)
        self.assertIsInstance(command, shelltools.Command)
        self.assertEqual(command.name, "echo")

    def test_check_output(self):
        with mock_patches(
            "ycyc.base.shelltools.subprocess.check_output",
        ) as patches:
            patches.check_output.return_value = "hello lyc"

            command = shelltools.Command("echo")

            self.assertEqual(
                command.check_output("hello", "lyc"),
                patches.check_output.return_value
            )
            call_args = patches.check_output.call_args[0]
            self.assertListEqual(call_args[0], ["echo", "hello", "lyc"])

    def test_check_call(self):
        with mock_patches(
            "ycyc.base.shelltools.subprocess.check_call",
        ) as patches:
            patches.check_call.return_value = 0

            command = shelltools.Command("echo")

            self.assertEqual(
                command.check_call("lyc"),
                patches.check_call.return_value
            )
            call_args = patches.check_call.call_args[0]
            self.assertListEqual(call_args[0], ["echo", "lyc"])

    def test_call(self):
        with mock_patches(
            "ycyc.base.shelltools.subprocess.Popen",
        ) as patches:
            command = shelltools.Command("echo")

            self.assertIs(
                command("hello", "lyc"),
                patches.Popen.return_value
            )

    def test_subprocessor(self):
        with mock_patches(
            "subprocess.Popen",
        ) as patches:
            command = shelltools.Command("echo")

            with command.subprocessor("hello", "lyc") as subprocessor:
                patches.Popen.assert_called_once_with(
                    ["echo", "hello", "lyc"],
                    stdout=shelltools.subprocess.PIPE,
                    stderr=shelltools.subprocess.PIPE,
                    shell=False,
                )



#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase
from collections import namedtuple
from textwrap import dedent

import mock

from ycyc.base import calctools

ObjModel = namedtuple("ObjModel", ["key"])


class TestSafeCalc(TestCase):
    def test_usage(self):
        calc = calctools.SafeCalc({
            "vint": 1, "vfloat": 2.3, "vstr": "456", "vlist": [7, 8.9, "10"],
            "vdict": {"key1": 11, "key2": ObjModel(key=12.13)}
        })
        self.assertEqual(calc("vint+1"), 2)
        self.assertEqual(calc("0.09 < vfloat/23 < 0.1"), True)
        self.assertEqual(calc("vstr == '456'"), True)
        self.assertEqual(calc("7 in vlist"), True)
        self.assertEqual(set(calc("vdict.keys()")), set(["key1", "key2"]))
        self.assertEqual(calc("vdict['key1']/10 == vint"), True)
        self.assertEqual(calc("12 and 34 or 56"), 34)

    def test_func(self):
        import operator

        calc = calctools.SafeCalc({
            "add": operator.add, "sub": operator.sub,
        })

        self.assertEqual(calc("add(1, sub(2, 3))"), 0)

        with self.assertRaises(calctools.NameParseError):
            calc("mul(4, 5)")

        with self.assertRaises(NameError):
            calc("mul")

        self.assert_(issubclass(calctools.NameParseError, NameError))

    def test_evil(self):
        forbidden_exprs = (
            "import os", "open('test.txt', 'w')", "def x(): pass; print x.func_globals",
            "().__class__.mro()[1].__subclasses__()"
        )
        calc = calctools.SafeCalc({})

        for e in forbidden_exprs:
            with self.assertRaises(Exception):
                print e
                calc(e)

    def test_timeout(self):
        import time

        calc = calctools.SafeCalc(
            {"sleep": time.sleep},
            timeout=0.05, interval=0.001,
        )
        try:
            calc("sleep(1)")
            self.assertTrue(False)
        except KeyboardInterrupt:
            pass
        except RuntimeError:
            pass

        with self.assertRaisesRegexp(SyntaxError, "invalid syntax"):
            calc("while True: pass")

    def test_allow_attr(self):
        m = mock.MagicMock()
        m.value = 1

        calc = calctools.SafeCalc({"model": m})
        self.assertEqual(calc("model.value + 1"), 2)

        calc = calctools.SafeCalc(
            {"model": m, "value": m.value}, allow_attr=False
        )
        with self.assertRaisesRegexp(NameError, "value not allow"):
            calc("model.value + 1")
        self.assertEqual(calc("value + 1"), 2)


class TesstSafeCalcFunc(TestCase):
    def test_usage(self):
        self.assertEqual(calctools.safecalc("1+1"), 2)
        self.assertEqual(calctools.safecalc("1+a", a=1), 2)
        self.assertEqual(calctools.safecalc("1+a", dict(a=1)), 2)
        self.assertEqual(calctools.safecalc("1+a", dict(a=1), a=2), 3)


class TestSelect(TestCase):
    def test_usage(self):
        obj = {
            "key": {"name": "test"},
            "value": [123],
        }
        self.assertDictEqual(
            calctools.select(obj, ("key", "name"), ("value", 0)),
            {0: obj["key"]["name"], 1: obj["value"][0]}
        )
        self.assertDictEqual(
            calctools.select(obj, key=("key", "name"), value=("value", 0)),
            {"key": obj["key"]["name"], "value": obj["value"][0]}
        )
        self.assertDictEqual(
            calctools.select(obj, noting=("key", "noting")),
            {"noting": None}
        )



from unittest import TestCase

from ycyc.ycollections.serial_list import SerialList


class TestSerialList(TestCase):
    def test_usage(self):
        serial_list = SerialList()
        self.assertEqual(0, serial_list.next_sn)
        serial_list.push(1, 1)
        serial_list.push(0, 2)
        serial_list.push(3, 3)

        item = serial_list.pop_item()
        self.assertEqual(0, item.sn)
        self.assertEqual(2, item.value)
        self.assertEqual(1, serial_list.next_sn)

        item = serial_list.pop_item()
        self.assertEqual(1, item.sn)
        self.assertEqual(1, item.value)
        self.assertEqual(2, serial_list.next_sn)

        item = serial_list.pop_item()
        self.assertIsNone(item)
        self.assertEqual(2, serial_list.next_sn)

        item = serial_list.pop_item(force=True)
        self.assertEqual(3, item.sn)
        self.assertEqual(3, item.value)
        self.assertEqual(4, serial_list.next_sn)

        serial_list.push(2, 4)
        item = serial_list.pop_item()
        self.assertIsNone(item)
        self.assertEqual(4, serial_list.next_sn)

        serial_list.push(4, 5)
        item = serial_list.pop_item()
        self.assertEqual(4, item.sn)
        self.assertEqual(5, item.value)
        self.assertEqual(5, serial_list.next_sn)
